﻿CompilerIf Defined(INCLUDE_HASHTABLECLASS, #PB_Constant)=0
#INCLUDE_HASHTABLECLASS=1
;/////////////////////////////////////////////////////////////////////////////////
;An OOP class for implementing structured hash tables.

;By Stephen Rodriguez.
;January 2010.

;Developed with Purebasic 4.4.
;Fully cross-platform.
;/////////////////////////////////////////////////////////////////////////////////
;
;/////////////////////////////////////////////////////////////////////////////////
;To create an instance of a new hash table class, use something like :
;
;   MyHashTable.HashTableObject
;   MyHashTable = NewHashTable(numberOfIndexes, sizeOfEachElement, numberOfStringFields [,threadsafeSwitch])
;
;and check the return value is non zero.
;
;Set the optional 'threadsafeSwitch' parameter to #True if the hash table is to be accessed
;by multiple threads. In which case the library will automatically protect shared access
;to the underlying data structures by using a mutex etc. Using the compiler threadsafe switch
;does not, in itself, protect access to shared resources by multiple threads!
;(Also compile with the PB threadsafe compiler switch of course!)
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;Methods exposed by the class :
;
;      AddItem(key$, userdata)
;      ClearItems()
;      Destroy()
;      EnumKeys()
;      GetItem.i(key$)
;      GetPointerToItem.i(key$)
;      IsKeyPresent.i(key$)
;      RemoveItem(key$)
;/////////////////////////////////////////////////////////////////////////////////
;
;/////////////////////////////////////////////////////////////////////////////////
;NOTE.
;=====
;This implementation of a hash table is more efficient than my previous one :
;   http://www.purebasic.fr/english/viewtopic.php?t=24683
;in that there is far less overhead per item stored. This is mainly because there is no 
;linked-list used to store the items resulting in far less data having to be stored to deal with
;possible 'collisions' etc.
;/////////////////////////////////////////////////////////////////////////////////

;-INCLUDES.
  XIncludeFile "HashTableClass_Residents.pbi"
  ;The following class for structured arrays is required  (http://www.purebasic.fr/english/viewtopic.php?t=29506)
    CompilerIf Defined(NewArray, #PB_Procedure) = 0
      XIncludeFile "..\array\arrayClass.pbi"
    CompilerEndIf

;-CONSTANTS.
  #HashTableClass_OVERWRITEDUPLICATES = #True ;Set to #FALSE if duplicates are to be ignored.


;-STRUCTURES.
  ;The following structure contains the class template and private properties,
    Structure _membersHashTableClass
      *vTable.i
      ptrTable.i
      numberOfIndexes.i
      mutex.i
      SizeOfEachElement.i
      NumberOfStringFields.i  ;All strng fields must reside at the beginning of the
                              ;structure.
    EndStructure 

  ;The following represents the minimum data stored for each item added to a hash table.
    Structure _HashTableItem
      key.s ;Text from which hash is calculated. Stored because of possible 'collisions'.
    EndStructure


;-PROTOTYPES.
  ;The following is used as part of the EnumKeys() method when calling the user's callback function.
    Prototype.i HashTableClass_ptyEnumKeys(key$, *ptr._HashTableItem, lParam)


;-DECLARES.
  Declare.i HashTableClass_FindKey(ArrayObject.StructuredArrayObject, key$, itemUBound, *ptrIndex.INTEGER)
  Declare.i HashTableClass_CreateHash(numelementsintable, *key.CHARACTER)



;/////////////////////////////////////////////////////////////////////////////////
;Create an instance of the HashTableClass.
;Returns zero if an error.
Procedure.i NewHashTable(numberOfHashes, SizeOfEachElement, NumberOfStringFields, threadsafeSwitch=0)
  Protected *object._membersHashTableClass
  If numberOfHashes
    ;Allocate memory for the object itself.
      *object = AllocateMemory(SizeOf(_membersHashTableClass))
      If *object
        ;Allocate memory for the table.
          *object\ptrTable = AllocateMemory(numberOfHashes*SizeOf(INTEGER))
          If *object\ptrTable
            ;Create a mutex if appropriate.
              If threadSafeSwitch
                *object\mutex = CreateMutex()
                If *object\mutex = 0
                  FreeMemory(*object\ptrTable)
                  FreeMemory(*object)
                  ProcedureReturn 0 ;Error.
                EndIf
              EndIf
              ;Set the object properties.
                *object\vTable = ?VTable_HashTableClass
                *object\numberOfIndexes = numberOfHashes
                *object\SizeOfEachElement = SizeOfEachElement
                *object\NumberOfStringFields = NumberOfStringFields
          Else
            FreeMemory(*object)
            *object = 0
          EndIf
      EndIf
  EndIf
  ProcedureReturn *object
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-CLASS METHODS.

;/////////////////////////////////////////////////////////////////////////////////
;The following method adds data to an existing hash table.
;If the key$ already exists, then we proceed as dictated by the value of #HashTableClass_OVERWRITEDUPLICATES.
;Returns zero if an error (this will be due to problems allocating a new structured array object) else returns a pointer
;to the newly added object.
Procedure.i HashTableClass_AddItem(*this._membersHashTableClass, *HTI._HashTableItem)
  Protected result, hash, *ptrInteger.INTEGER, ArrayObject.StructuredArrayObject, htItem._HashTableItem
  Protected itemUBound, blnKeyAlreadyExist, itemPos
  ;First job is to create the hash index.
    hash = HashTableClass_CreateHash(*this\numberOfIndexes-1, @*HTI\key)
  ;Check if there are already entries made with this same hash.
    *ptrInteger = *this\ptrTable+hash*SizeOf(INTEGER)
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  If *ptrInteger\i = 0 ;This is the first entry with this hash. We thus create a
                    ;new structured array object to hold these values.
    ArrayObject = NewStructuredArray(0, *this\SizeOfEachElement, *this\NumberOfStringFields)
    If ArrayObject
      *ptrInteger\i = ArrayObject
      result = ArrayObject\SetValue(0, *HTI)
    Else
      result = 0 ;Error.
    EndIf
  Else ;Either the key already exists or we have a collision.
    ArrayObject = *ptrInteger\i
    ;We must search the array looking for either the position to insert the new item (so that
    ;we can perform binary searches) or an item with the same key etc.
      itemUBound = ArrayObject\GetUpperBound()
      blnKeyAlreadyExist = HashTableClass_FindKey(ArrayObject, *HTI\key, itemUBound, @itemPos)
      If blnKeyAlreadyExist ;Collision.
        ;We overwrite this element only if #HashTableClass_OVERWRITEDUPLICATES = #True.
          If #HashTableClass_OVERWRITEDUPLICATES
            result = ArrayObject\SetValue(itemPos, *HTI)
          EndIf
      Else ;There is no entry sharing the same key.
        ;First add a new entry to the structured array.
          If ArrayObject\ReDim(itemUBound+1)
            ;Shift elements right to make room for the new entry.
              ArrayObject\ShiftElementsRight(itemPos)
            ;Add the new entry.
              result = ArrayObject\SetValue(itemPos, *HTI)
          EndIf
      EndIf
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method removes all items from the underlying hash table.
Procedure HashTableClass_ClearItems(*this._membersHashTableClass)
  Protected i, *ptrInteger.INTEGER, ArrayObject.StructuredArrayObject
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  *ptrInteger = *this\ptrTable
  ;Remove the structured arrays used to store the data.
    For i = 0 To *this\numberOfIndexes-1
      If *ptrInteger\i ;There are entries here to remove.
        ArrayObject = *ptrInteger\i
        ArrayObject\Destroy()  ;All elements are automatically freed by the structured array class.
        *ptrInteger\i = 0
      EndIf
      *ptrInteger + SizeOf(INTEGER)
    Next
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method frees all memory associated with the underlying hash table.
Procedure HashTableClass_Destroy(*this._membersHashTableClass)
  Protected i, *ptrInteger.INTEGER, ArrayObject.StructuredArrayObject
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  *ptrInteger = *this\ptrTable
  ;First remove the structured arrays used to store the data.
    For i = 0 To *this\numberOfIndexes-1
      If *ptrInteger\i ;There are entries here to remove.
        ArrayObject = *ptrInteger\i
        ArrayObject\Destroy()  ;All elements are automatcally freed by the structured array class.
      EndIf
      *ptrInteger + SizeOf(INTEGER)
    Next
  ;Now free the table of pointers.
    FreeMemory(*this\ptrTable)
  ;Now any mutex.
    If *this\mutex
      UnlockMutex(*this\mutex)
      FreeMutex(*this\mutex)
    EndIf
  ;Now the object itself.
    FreeMemory(*this)
    *this = 0
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following enumerates through all the keys in the underlying hash table (in reverse)
;and passes the key$ and a pointer to the underlying structure to the user defined callback.
;The user's callback must return #False to discontinue the enumeration.
;'lParam' is an application defined value which is also passed to the callback.
Procedure HashTableClass_EnumKeys(*this._membersHashTableClass, addrCallback, lParam=0)
  Protected result, i, fn.HashTableClass_ptyEnumKeys, *ptrInteger.INTEGER, ArrayObject.StructuredArrayObject
  Protected *htItem._HashTableItem
  If addrCallback
    fn = addrCallback
    If *this\mutex
      LockMutex(*this\mutex)
    EndIf
    *ptrInteger = *this\ptrTable + (*this\numberOfIndexes - 1) * SizeOf(INTEGER) 
    Repeat
      If *ptrInteger\i  ;There are entries with this hash.
        ArrayObject = *ptrInteger\i
        For i = ArrayObject\GetUpperBound() To 0 Step -1 ;Reverse order in case the user's callback starts removing items.
          ;In case elements are being removed or added during the enumeration.
            ArrayObject = *ptrInteger\i 
            If i>ArrayObject\GetUpperBound()
              i = ArrayObject\GetUpperBound()
            EndIf
          *htItem = ArrayObject\GetPointerToValue(i)
          ;Call the user's callback.
            result = fn(*htItem\key, *htItem, lParam)
          ;Has the user signalled that we should discontinue?
          If result = #False
            Break 2
          EndIf
        Next
      EndIf
      *ptrInteger - SizeOf(INTEGER)
    Until *ptrInteger < *this\ptrTable
    If *this\mutex
      UnlockMutex(*this\mutex)
    EndIf
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function retrieves data from the underlying hash table.
;*HTI must point to a structure whose fields will be filled from the hash table.
;When this function is called, the key field of *HTI must contain the key of the item
;being sought.
;Returns zero if an item with the specified key does not exist in the hash table.
;You can use the IsKeyPresent() method to determine beforehand if the key exists etc.
Procedure.i HashTableClass_GetItem(*this._membersHashTableClass, *HTI._HashTableItem)
  Protected result, hash, *ptrInteger.INTEGER, ArrayObject.StructuredArrayObject
  Protected blnKeyAlreadyExist, itemPos, itemUBound
  ;First job is to create the hash index.
    hash = HashTableClass_CreateHash(*this\numberOfIndexes-1, @*HTI\key)
    *ptrInteger = *this\ptrTable+hash*SizeOf(INTEGER)
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  ;Check if there are already entries made with this same hash.
    If *ptrInteger\i  ;There are entries with this hash. Can we find one with a matching key?
      ArrayObject = *ptrInteger\i
      ;We must search the array looking for item with the same key etc.
        itemUBound = ArrayObject\GetUpperBound()
        blnKeyAlreadyExist = HashTableClass_FindKey(ArrayObject, *HTI\key, itemUBound, @itemPos)
        If blnKeyAlreadyExist ;Found!
          ArrayObject\GetValue(itemPos, *HTI)
          result = 1
        EndIf
    EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following function retrieves a pointer to the data from the underlying hash table.
;Returns zero if an item with the specified key does not exist in the hash table.
;You can use the IsKeyPresent() method to determine beforehand if the key exists etc.
Procedure.i HashTableClass_GetPointerToItem(*this._membersHashTableClass, key.s)
  Protected result, hash, *ptrInteger.INTEGER, ArrayObject.StructuredArrayObject
  Protected blnKeyAlreadyExist, itemPos, itemUBound
  ;First job is to create the hash index.
    hash = HashTableClass_CreateHash(*this\numberOfIndexes-1, @key)
    *ptrInteger = *this\ptrTable+hash*SizeOf(INTEGER)
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  ;Check if there are already entries made with this same hash.
    If *ptrInteger\i  ;There are entries with this hash. Can we find one with a matching key?
      ArrayObject = *ptrInteger\i
      ;We must search the array looking for item with the same key etc.
        itemUBound = ArrayObject\GetUpperBound()
        blnKeyAlreadyExist = HashTableClass_FindKey(ArrayObject, key, itemUBound, @itemPos)
        If blnKeyAlreadyExist ;Found!
          result = ArrayObject\GetPointerToValue(itemPos)
        EndIf
    EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following function checks a key value to see if it exists within the underlying hash table.
;A non-zero return indicates that the key is present within the hash table.
Procedure.i HashTableClass_IsKeyPresent(*this._membersHashTableClass, key.s)
  Protected result, hash, *ptrInteger.INTEGER, ArrayObject.StructuredArrayObject
  Protected blnKeyAlreadyExist, itemPos, itemUBound
  ;First job is to create the hash index.
    hash = HashTableClass_CreateHash(*this\numberOfIndexes-1, @key)
    *ptrInteger = *this\ptrTable+hash*SizeOf(INTEGER)
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  If *ptrInteger\i  ;There are entries with this hash. Can we find one with a matching key?
    ArrayObject = *ptrInteger\i
    ;We must search the array looking for item with the same key etc.
      itemUBound = ArrayObject\GetUpperBound()
      blnKeyAlreadyExist = HashTableClass_FindKey(ArrayObject, key, itemUBound, @itemPos)
      If blnKeyAlreadyExist ;Found!
        result = 1
      EndIf
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following function removes an entry from an existing hash table.
Procedure HashTableClass_RemoveItem(*this._membersHashTableClass, key.s)
  Protected hash, *ptrInteger.INTEGER, ArrayObject.StructuredArrayObject
  Protected blnKeyAlreadyExist, itemPos, itemUBound
  ;First job is to create the hash index.
    hash = HashTableClass_CreateHash(*this\numberOfIndexes-1, @key)
    *ptrInteger = *this\ptrTable+hash*SizeOf(INTEGER)
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  If *ptrInteger\i  ;There are entries with this hash. Can we find one with a matching key?
    ArrayObject = *ptrInteger\i
    ;We must search the array looking for item with the same key etc.
      itemUBound = ArrayObject\GetUpperBound()
      blnKeyAlreadyExist = HashTableClass_FindKey(ArrayObject, key, itemUBound, @itemPos)
      If blnKeyAlreadyExist ;Found!
        If itemUBound = 0
          ;Delete the array.
            ArrayObject\Destroy()
            *ptrInteger\i = 0
        Else
          ;We need to shift the structured array elements down in order to remove this item.
          ;All the memory freeing is taken care of by the structured array class.
            ArrayObject\ShiftElementsLeft(itemPos)
          ;Redimension.
            ArrayObject\ReDim(itemUBound-1)
        EndIf
      EndIf
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-INTERNAL FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
;The following performs a binary search on a structured array object used to hold
;all the items having a particular hash (collisions).
;Returns #True if the key already exists.
;In either case, the index of the appropriate position is placed in the variable pointed
;to by '*ptrIndex'.
Procedure.i HashTableClass_FindKey(ArrayObject.StructuredArrayObject, key$, itemUBound, *ptrIndex.INTEGER)
  Protected low, high, ptr, *htItem._HashTableItem, blnFound
  low = 0 : high = itemUBound : ptr = high/2
  Repeat 
    *htItem = ArrayObject\GetPointerToValue(ptr)
    If *htItem\key = key$
      *ptrIndex\i = ptr
      ProcedureReturn 1 ;Key found!
    ElseIf *htItem\key < key$
      If low < ptr
        low = ptr
      Else
        low = ptr+1
      EndIf
    Else
      If high>ptr
        high = ptr    
      Else
        low=ptr-1
        If low<0
          low=0
        EndIf
      EndIf
    EndIf
    ptr = (low + high)/2
  Until low>=high
  *htItem = ArrayObject\GetPointerToValue(ptr)
  If *htItem\key = key$
    *ptrIndex\i = ptr
    ProcedureReturn 1 ;Key found!
  ElseIf *htItem\key < key$
    *ptrIndex\i = ptr + 1
    ProcedureReturn 0 ;Key not found!
  Else
    *ptrIndex\i = ptr
    ProcedureReturn 0 ;Key not found!
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-HASH FUNCTION
;/////////////////////////////////////////////////////////////////////////////////
;The following function creates the hash code.
;This can be altered to use a different alogorithm.
;The function MUST return a value between 0 and 'numelementsintable'.
;We currently use the 'djb2' algorithm.
;You do not need to worry about collisions, the library automatically takes care of this.

;NOTE, the following returns the same code when working in either ASCII or UNICODE.
;Probably best to ensure that any replacement algorithm does likewise, so avoid PeekB() etc.
;Either use PeekC() (and take care with pointers) or Asc().
Procedure.i HashTableClass_CreateHash(numelementsintable, *key.CHARACTER)
  Protected hash=5381 
  While *key\c <> 0 
    hash=(hash<<5)+hash + *key\c ; hash * 33 + asc 
    *key+SizeOf(CHARACTER)
  Wend 
  If hash>=0 
    ProcedureReturn hash%(numelementsintable+1) 
  Else 
    ProcedureReturn -(hash%(numelementsintable+1)) 
  EndIf 
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-VIRTUAL TABLES.

DataSection 
  VTable_HashTableClass: 
    Data.i @HashTableClass_AddItem() 
    Data.i @HashTableClass_ClearItems()
    Data.i @HashTableClass_Destroy()
    Data.i @HashTableClass_EnumKeys()
    Data.i @HashTableClass_GetItem()
    Data.i @HashTableClass_GetPointerToItem()
    Data.i @HashTableClass_IsKeyPresent()
    Data.i @HashTableClass_RemoveItem()
EndDataSection 
CompilerEndIf

; IDE Options = PureBasic 4.40 (Windows - x86)
; CursorPosition = 241
; FirstLine = 241
; Folding = --
; Markers = 242